#!/usr/bin/env python
#
#   zfilenet/simpleclient.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Simple command-line client for ZFN.
"""

from __future__ import absolute_import
from base64 import b64decode, b64encode
from hashlib import sha1
import logging
import os
import sys

import pinet
from pinet.dispatch import expose
import progtools
from twisted.internet import reactor

from . import discovery, indexserver, util

__author__ = 'Ross Light'
__date__ = 'June 15, 2008'
__version__ = '0.2.0b2'
__all__ = ['usage',
           'description',
           'log',
           'ClientInterface',
           'ClientProtocol',
           'ClientFactory',
           'main',]

usage = "usage: %prog host:port <ls|stat|mkdir|upload|download|mv|cp|rm>"
description = "Command-line access to ZFN JSON-RPC API"
log = logging.getLogger('zfilenet.client.simple')

class ClientInterface(object):
    def __init__(self):
        self.dreceive = util.ProgressDeferred()
        self.dfinish = util.ProgressDeferred()
    
    @expose('download.receive_data')
    def download_receive_data(self, file_id, data):
        file_id = util.to_uuid(file_id)
        data = b64decode(data)
        log.debug("Received %i bytes for %s", len(data), file_id)
        self.dreceive.callback((file_id, data))
    
    @expose('download.close')
    def download_close(self, file_id, success):
        file_id = util.to_uuid(file_id)
        log.debug("Finished download for %s (success: %s)", file_id, success)
        self.dfinish.callback((file_id, success))

class ClientProtocol(pinet.jsonrpc.JSON_RPCProtocol):
    def connectionMade(self):
        pinet.jsonrpc.JSON_RPCProtocol.connectionMade(self)
        self.factory.do_callback()

class ClientFactory(pinet.jsonrpc.JSON_RPCFactory):
    protocol = ClientProtocol
    
    def __init__(self, action, *args, **kw):
        pinet.jsonrpc.JSON_RPCFactory.__init__(self)
        self.action = action
        self.callbackArgs, self.callbackKw = args, kw
    
    def clientConnectionFailed(self, connector, reason):
        if reactor.running:
            log.critical("Connection failed (%s)", reason.value)
            reactor.stop()
    
    def clientConnectionLost(self, connector, reason):
        if reactor.running:
            log.critical("Connection lost (%s)", reason.value)
            reactor.stop()
    
    def do_callback(self):
        self.action(self, *self.callbackArgs, **self.callbackKw)

def error_handler(failure):
    log.error("%s", failure.value.message)
    reactor.stop()

def react_usage(message):
    log.log(util.LOG_USAGE, "%s: usage: %s",
            progtools.get_program_name(), message)
    reactor.stop()

def run_stat(factory, options, args, peer):
    # Display settings
    key_order = ['uuid',
                 'type',
                 'created',
                 'modified',
                 'accessed',
                 'length',
                 'sha1_hash',]
    labels = {'uuid': "UUID",
              'type': "Object Type",
              'created': "Created",
              'modified': "Last Modified",
              'accessed': "Last Accessed",
              'length': "File Size",
              'sha1_hash': "SHA1 Checksum",}
    label_field_size = max(len(label) for label in labels.itervalues())
    converters = {'length': util.format_file_size,}
    # Callback
    def _got_stat(data):
        if data is None:
            log.error("Path does not exist")
        else:
            for key in key_order:
                # Fetch value
                value = data.get(key)
                if value is None:
                    continue
                # Convert value (if needed)
                try:
                    func = converters[key]
                except KeyError:
                    pass
                else:
                    value = func(value)
                # Display value
                label = labels[key]
                print "%*s: %s" % (label_field_size, label, value)
        reactor.stop()
    # Kick off process
    if len(args) != 1:
        react_usage("stat takes one argument")
        return
    output_file = progtools.open_output_file(options.output_file)
    d = factory.call('stat', args[0])
    d.addCallback(_got_stat)
    d.addErrback(error_handler)

def run_mkdir(factory, options, args, peer):
    def _finished_mkdir(result):
        reactor.stop()
    # Kick off process
    if len(args) != 1:
        react_usage("mkdir takes one argument")
        return
    d = factory.call('mkdir', args[0])
    d.addCallback(_finished_mkdir)
    d.addErrback(error_handler)

def run_mv(factory, options, args, peer):
    def _finished_mv(result):
        reactor.stop()
    # Kick off process
    if len(args) != 2:
        react_usage("mv takes two arguments")
        return
    d = factory.call('move', args[0], args[1])
    d.addCallback(_finished_mv)
    d.addErrback(error_handler)

def run_cp(factory, options, args, peer):
    def _finished_cp(result):
        reactor.stop()
    # Kick off process
    if len(args) != 2:
        react_usage("cp takes two arguments")
        return
    d = factory.call('copy', args[0], args[1])
    d.addCallback(_finished_cp)
    d.addErrback(error_handler)

def run_rm(factory, options, args, peer):
    def _finished_rm(result):
        reactor.stop()
    # Kick off process
    if len(args) != 1:
        react_usage("rm takes one argument")
        return
    d = factory.call('delete', args[0])
    d.addCallback(_finished_rm)
    d.addErrback(error_handler)

def run_ls(factory, options, args, peer):
    def _got_listing(children):
        if children is None:
            log.error("Directory does not exist")
        else:
            for name in sorted(children):
                print >> output_file, name
        reactor.stop()
    # Kick off process
    if len(args) != 1:
        react_usage("ls takes one argument")
        return
    output_file = progtools.open_output_file(options.output_file)
    d = factory.call('listdir', args[0])
    d.addCallback(_got_listing)
    d.addErrback(error_handler)

class Downloader(object):
    def __init__(self, out_path, d, peer):
        self.output_file = progtools.open_output_file(out_path, 'wb')
        self.file_hash = sha1()
        self.file_length = 0
        d.addCallback(self.got_info)
        d.addErrback(self.got_error)
        peer.dreceive.addCallback(self.got_data)
        peer.dfinish.addCallback(self.finished_download)
    
    def got_info(self, result):
        log.debug("Got info: %r", result)
        self.file_info = result
        self.file_id = util.to_uuid(self.file_info['file_id'])
    
    def got_error(self, failure):
        log.error("Download error: %s", failure.value.message)
        reactor.stop()
    
    def got_data(self, result):
        file_id, data = result
        if file_id != self.file_id:
            log.warning("Received data for wrong file (UUID %s)", file_id)
        else:
            self.file_hash.update(data)
            self.file_length += len(data)
            self.output_file.write(data)
    
    def finished_download(self, result):
        file_id, success = result
        if file_id != self.file_id:
            log.warning("Received data for wrong file (UUID %s)", file_id)
            return
        final_hash = self.file_hash.hexdigest()
        if not success:
            log.error("Download failed (a blob is missing).")
        elif self.file_length != self.file_info['file_length']:
            log.error("Download error: File length does not match server "
                      "(got %i, expected %i)",
                      final_size, self.file_info['file_length'])
        elif final_hash != self.file_info['file_hash']:
            log.error("Download error: File checksum invalid")
        else:
            log.info("Download successful!")
        self.output_file.close()
        reactor.stop()

def run_download(factory, options, args, peer):
    # Get arguments
    if len(args) != 1:
        react_usage("download takes one argument")
        return
    # Kick off process
    d = factory.call('download.open', args[0])
    Downloader(options.output_file, d, peer)

def run_upload(factory, options, args, peer):
    UPLOAD_SIZE = 128 * 1024
    # Callbacks
    def _upload_failed(failure):
        log.error("Upload error: %s", failure.value.message)
        reactor.stop()
    def _started_upload(result):
        file_id = util.to_uuid(result['upload_id'])
        lock = result['lock']
        _send_data(file_id, lock)
    def _sent_data(result, file_id, lock):
        _send_data(file_id, lock)
    def _finished_upload(result, file_id):
        log.info("Upload successful!")
        reactor.stop()
    # The real McCoy upload handler
    def _send_data(file_id, lock):
        data = input_file.read(UPLOAD_SIZE)
        if data:
            log.debug("Sending %i bytes...", len(data))
            d = factory.call('upload.write', file_id.hex, b64encode(data),
                             lock)
            d.addCallback(_sent_data, file_id, lock)
            d.addErrback(_upload_failed)
        else:
            input_file.close()
            d = factory.call('upload.close', file_id.hex, lock)
            d.addCallback(_finished_upload, file_id)
            d.addErrback(_upload_failed)
    # Get arguments
    if len(args) == 1:
        server_destination, input_file = args[0], '-'
    elif len(args) == 2:
        server_destination, input_file = args[1], args[0]
    else:
        react_usage("upload [input-file] destination")
        return
    # Kick off process
    input_file = progtools.open_input_file(input_file, 'rb')
    d = factory.call('upload.open', server_destination)
    d.addCallback(_started_upload)
    d.addErrback(_upload_failed)

@progtools.reportexits
def main(args=None):
    # Parse arguments
    if args is None:
        args = sys.argv[1:]
    parser = progtools.OptionParser(usage=usage, version=__version__,
                                    description=description,)
    parser.set_defaults(output_file='-',)
    util.add_log_options(parser)
    parser.add_option('-o', '--output',
                      action='store',
                      dest='output_file',
                      metavar="FILE",
                      help="Changes the port the server runs on. "
                           "[default: stdout]")
    options, args = parser.parse_args(args)
    util.setup_log_options(options)
    if len(args) == 0:
        parser.print_help()
        return 0
    elif len(args) == 1:
        raise progtools.UsageError()
    else:
        host, action, args = args[0], args[1], args[2:]
        try:
            host, port = host.rsplit(':')
        except ValueError:
            # No port given, use default
            host, port = host, indexserver.rpcapi.DEFAULT_PORT
        else:
            # Parse port
            port = int(port, 10)
    # Get action
    actions = {'stat': run_stat,
               'ls': run_ls,
               'mkdir': run_mkdir,
               'rm': run_rm,
               'mv': run_mv,
               'cp': run_cp,
               'download': run_download,
               'upload': run_upload,}
    if action in actions:
        action = actions[action]
    else:
        log.error("Invalid action given: %s", action)
        return 1
    # Set up client
    peer = ClientInterface()
    factory = ClientFactory(action, options, args, peer)
    factory.requests.register_object(peer)
    # Run reactor
    log.debug("Connecting to zfn://%s:%i/...", host, port)
    reactor.connectTCP(host, port, factory)
    log.debug("Running reactor...")
    reactor.run()
    return 0

if __name__ == '__main__':
    sys.exit(main())
